from __future__ import annotations
import asyncio
import json
from asyncio import Future
from pathlib import Path
from typing import List, Type, TypeVar, Generic, Dict, Any

from dependency_injector import providers, containers, resources
from dependency_injector.wiring import inject, Provide
from pydantic import BaseSettings

from web_foundation.kernel.dispatcher import IDispatcher
from web_foundation.app.resources.cache.cache import AppCache
from web_foundation.app.resources.cache.memory import InMemoryCache
from web_foundation.app.service import Service
from web_foundation.kernel.logger import setup_loggers, LoggerSettings
from web_foundation.workers.worker import GenWorker

AppConfig = TypeVar("AppConfig", bound=BaseSettings)


class App(Generic[AppConfig, GenWorker]):
    config: AppConfig
    name: str
    workers: Dict[str, GenWorker]
    services: Dict[str, Service]
    dispatcher: IDispatcher
    debug: bool  # type: ignore

    def __init__(self, config: providers.Configuration, debug: bool = False):
        self.debug = debug
        self.name = config.get("app_name")
        self.config = config
        self.workers = {}
        self.services = {}
        self.dispatcher = IDispatcher()

    @inject
    async def _before_app_run(self, app_container: AppContainer = Provide["<container>"]):
        await app_container.init_resources()
        await self.before_app_run(app_container)
        await app_container.shutdown_resources()

    async def before_app_run(self, container):
        pass

    def _add_worker(self, worker: GenWorker):
        if not worker.configured:
            raise RuntimeError("Worker not configured")
        worker.debug = self.debug
        self.dispatcher.reg_worker(worker)
        self.workers.update({worker.name: worker})

    def add_worker(self, worker: GenWorker | List[GenWorker]):
        """
        Set isolate to app and set isolate debug and name
        :param worker: isolate to apply
        :return: None
        """
        if isinstance(worker, list):
            for w in worker:
                self._add_worker(w)
        else:
            self._add_worker(worker)

    def add_service(self, service: Service):
        self.services.update({service.__class__.__name__: service})

    def find_worker_by_pid(self, pid: int):
        for worker in self.workers.values():
            if worker.pid == pid:
                return worker

    def wire_app(self, container: AppContainer):
        container.wire(modules=[self.__module__])
        for worker in self.workers.values():
            container.wire(modules=[worker.__module__])
            worker.wire_worker(container)

    def perform(self) -> List[Future[Any]]:
        """
        Call performs from isolates and return all isolates Futures
        :return: None
        """
        return [worker.perform() for worker in self.workers.values()]

    async def run(self):
        await self._before_app_run()
        """
        Func to run app manually (without Runner)
        :return: None
        """
        return await asyncio.wait(self.perform() + self.dispatcher.preform())


def load_config(conf_path: Path, config_model: Type[AppConfig]) -> AppConfig:
    """
    Load app config to user in
    :param conf_path:
    :param config_model: BaseModel to cast json file to pydantic
    :return: None
    """
    with open(conf_path, "r") as _json_file:
        conf = config_model(**json.loads(_json_file.read()))
        return conf


class AppContainer(containers.DeclarativeContainer):
    debug = True
    app_config = providers.Configuration()
    app = providers.Dependency(instance_of=App, default=providers.Singleton(App, config=app_config, debug=debug))
    cache = providers.Dependency(instance_of=AppCache, default=providers.Singleton(InMemoryCache))
    """
    Please pass here your database or any another resource in nested container
    """

    def apply_resource(self, name: str, resource: Type[resources.AsyncResource], **kwargs):
        setattr(self, name, providers.Resource(resource, app_name=self.app_config.get("app_name"), **kwargs))



import sys
from traceback import extract_tb

import loguru
from sanic.errorpages import exception_response, JSONRenderer
from sanic.handlers import ErrorHandler
from sanic.exceptions import SanicException

from web_foundation.app.errors.application.application import ApplicationError


class StatusCodes:
    APPLICATION_CODE = 409
    REPORT_CODE = 501


class ExtJSONRenderer(JSONRenderer):
    def _generate_output(self, *, full):
        output = {
            "description": self.title,
            "status": self.status,
            # "message": self.text,
            "error": getattr(self.exception, "context", {})
        }
        if not output['error'].get('type'):
            output['error']['type'] = getattr(self.exception, "error_type", "C")

        if not output['error'].get('class'):
            output['error']['class'] = getattr(self.exception, "default_class", 999)

        if not output['error'].get('subclass'):
            output['error']['subclass'] = getattr(self.exception, "default_subclass", 999)

        if not output['error'].get('code'):
            output['error'][
                'code'] = f"{output['error']['type']}-{output['error']['class']}-{output['error']['subclass']}"

        if self.text:
            if output.get('error'):
                if not output['error'].get('comment'):
                    output['error'].update({"comment": self.text})
            else:
                output['error'] = {"comment": self.text}

        if full:
            _, exc_value, __ = sys.exc_info()
            exceptions = []

            while exc_value:
                exceptions.append(
                    {
                        "type": exc_value.__class__.__name__,
                        "exception": str(exc_value),
                        "frames": [
                            {
                                "file": frame.filename,
                                "line": frame.lineno,
                                "name": frame.name,
                                "src": frame.line,
                            }
                            for frame in extract_tb(exc_value.__traceback__)
                        ],
                    }
                )
                exc_value = exc_value.__cause__

            output["path"] = self.request.path
            output["args"] = self.request.args
            output["exceptions"] = exceptions[::-1]

        return output


class ExtendedErrorHandler(ErrorHandler):
    def default(self, request, exception):
        """Convert ApplicationError to SanicException"""
        renderer = None
        if isinstance(exception, ApplicationError):
            if not isinstance(exception, SanicException):
                exception = type(exception.__class__.__name__,
                                 (SanicException,),
                                 {"message": exception.message if exception.message else "",
                                  "status_code": StatusCodes.APPLICATION_CODE,
                                  "error_type": exception.error_type,
                                  "default_class": exception.default_class,
                                  "default_subclass": exception.default_subclass, })(context=exception.context)
            renderer = ExtJSONRenderer
        else:
            self.log(request, exception)
        fallback = request.app.config.FALLBACK_ERROR_FORMAT
        return exception_response(
            request,
            exception,
            debug=self.debug,
            base=self.base,
            fallback=fallback,
            renderer=renderer
        )



from __future__ import annotations

import io
import json
from dataclasses import dataclass, field
from typing import Any
from typing import Dict

from sanic.response import BaseHTTPResponse

from web_foundation.workers.io.http.chaining import InputContext
from web_foundation.workers.io.rt_connection import WriteableObj, RtConnection, RtMessage


@dataclass
class SseRtMessage(RtMessage):
    _SEPARATOR = "\r\n"
    event_name: str
    event_id: str | None = field(default=None)
    data: Dict | None = field(default=None)
    retry: int | None = field(default=None)

    def _prepare(self, *args, **kwargs) -> str | bytes:
        buffer = io.StringIO()
        buffer.write('event: ' + self.event_name + self._SEPARATOR)
        if self.event_id:
            buffer.write('id: ' + self.event_id + self._SEPARATOR)
        if self.retry:
            buffer.write('retry: ' + str(self.retry) + self._SEPARATOR)
        else:
            buffer.write("retry: " + "0" + self._SEPARATOR)
        if self.data:
            buffer.write('data: ' + json.dumps(self.data))
        else:
            buffer.write('data: ')
        buffer.write("\r\n\r\n")
        return buffer.getvalue()

    @classmethod
    def ping_message(cls):
        return cls(event_id="0", event_name="ping").to_sent


class WriteableSse(WriteableObj):
    obj: BaseHTTPResponse

    def __init__(self, obj: BaseHTTPResponse):
        super().__init__(obj)

    async def write(self, message: Any) -> None:
        await self.obj.send(message)


class SseRtConnection(RtConnection):
    writeable: WriteableSse

    def _ping_msg(self) -> str | bytes:
        return SseRtMessage.ping_message()

    @classmethod
    async def accept_connection(cls, input_ctx: InputContext, ping_enable: bool = True, ping_timeout=None,
                                listen_timeout=None, debug: bool = True) -> SseRtConnection:
        writable = await cls._construct_writable(input_ctx)
        instance = cls(input_ctx, writable, ping_enable, ping_timeout, listen_timeout, debug)
        return instance

    @classmethod
    async def _construct_writable(cls, input_ctx: InputContext, *args, **kwargs) -> WriteableSse:
        _response = await input_ctx.request.respond(content_type="text/event-stream")
        return WriteableSse(_response)

    def _close_condition(self) -> bool:
        if not self.writeable.obj.stream:
            return True
        else:
            return False

# class SseConnection(Generic[GenericIMessage]):
#     ping_timeout: float
#     listen_timeout: float
#     input_ctx: InputContext
#     connection_queue: Queue[GenericIMessage]
#     _last_event_id: int
#     _response: HTTPResponse
#     _alive: bool
#
#     def __init__(self, input_ctx, connection_queue: Queue[GenericIMessage], ping_timeout=None, listen_timeout=None):
#         self.input_ctx = input_ctx
#         self.connection_queue = connection_queue
#         self.ping_timeout = ping_timeout if ping_timeout else 5
#         self.listen_timeout = listen_timeout if listen_timeout else 0.1
#         self._last_event_id = 0
#
#     async def stream_queued_events(self, resolve_callback: ResolvedCallback) -> None:
#         async def _ping(ping_timeout: float):
#             ping_event = await SseProto.ping_message()
#             while True:
#                 if not self._response.stream:
#                     return
#                 await asyncio.sleep(ping_timeout)
#                 loguru.logger.warning("PINFFFFFF")
#                 await self._response.send(ping_event)
#
#         self._response = await self.input_ctx.request.respond(content_type="text/event-stream")
#         ping_task = self.input_ctx.request.app.loop.create_task(_ping(self.ping_timeout))
#         while True:
#             if not self._response.stream:
#                 ping_task.cancel()
#                 print(f"Sse on {self.input_ctx.request.ip} disconnected, ping task canceled")
#                 break
#             if not self.connection_queue.empty():
#                 resolved = await resolve_callback(self, await self.connection_queue.get())
#                 if resolved.disconnect:
#                     break
#                 if resolved.event_data and resolved.event_name != "":
#                     ev_payload = SseProto(event_id=str(self._last_event_id), event_name=resolved.event_name,
#                                           data=resolved.event_data).to_send
#                     await self._response.send(ev_payload.encode())
#                     self._last_event_id += 1
#             await asyncio.sleep(self.listen_timeout)


# ResolvedCallback = Callable[[SseConnection, GenericIMessage], Coroutine[Any, Any, SseResolved]]



import importlib
import os
from pathlib import Path
from types import ModuleType
from typing import List, Optional, Dict

from dependency_injector import resources
from loguru import logger
from pydantic import BaseModel as PDModel
from tortoise import Tortoise, BaseDBAsyncClient, connections


class DbConfig(PDModel):
    host: str
    port: str
    database: str
    user: str
    password: str
    db_schema: str
    with_migrations: bool
    migrations_path: Path


class DatabaseResource(resources.AsyncResource):
    modules: List[str]
    connection: BaseDBAsyncClient | None
    _app_name: str

    def _get_connection_setting(self) -> dict:
        if self.db_conf.get("with_migrations"):
            self.modules.append("aerich.models")
        to_discover = [importlib.import_module(i) for i in self.modules]
        return {
            'connections': {
                # Dict format for connection
                f'{self._app_name}_default': {
                    'engine': self.engine,
                    'credentials': {
                        'host': self.db_conf['host'],
                        'port': self.db_conf['port'],
                        'user': self.db_conf["user"],
                        'password': self.db_conf["password"],
                        'database': self.db_conf["database"],
                        'schema': self.db_conf["db_schema"],
                        'minsize': 1,
                        'maxsize': 5,
                    }
                }
            },
            'apps': {
                f'{self._app_name}': {
                    'models': to_discover,
                    'default_connection': f'{self._app_name}_default',
                }
            },
            'use_tz': False,
            'timezone': 'UTC'
        }

    async def fill_db_data(self):
        pass

    async def init(self, app_name: str, db_config: Dict, modules: List[str],
                   engine: str = 'tortoise.backends.asyncpg') -> BaseDBAsyncClient:
        super().__init__()
        self._app_name = app_name
        self.db_conf = db_config
        self.modules = modules
        self.connection = None
        self.engine = engine
        await Tortoise.init(config=self._get_connection_setting())
        await self.configure_db()
        return Tortoise.get_connection(f'{self._app_name}_default')

    async def shutdown(self, resource: Optional[Tortoise]) -> None:
        await self.close()

    @property
    def conn(self) -> BaseDBAsyncClient:
        if not self.connection:
            self.connection = Tortoise.get_connection(connection_name=f'{self._app_name}_default')
        return self.connection

    async def _migrations(self, schema_exists: bool, command):
        path_exists = os.path.exists(os.path.join(os.getcwd(), self.db_conf["migrations_path"]))

        if not path_exists and not schema_exists:
            await self.create_schema(False)
            await command.init()
            await command.init_db(safe=True)
        elif not schema_exists:
            await self.create_schema(False)
            await command.init()
            await command.upgrade()
        else:
            await command.init()
        logger.info(f"Apply migrations from {self.db_conf['migrations_path']}")
        await command.migrate()
        await command.upgrade()

    async def configure_db(self):
        scheme_name = self.db_conf.get('db_schema')
        row_count, rows = await self.conn.execute_query(
            f"SELECT schema_name FROM information_schema.schemata WHERE schema_name = '{scheme_name}'")
        schema_exists = True if row_count else False
        if self.db_conf.get('with_migrations'):
            if not self.db_conf['migrations_path']:
                raise ValueError("Field needmigrations_aeirch in db config set to false, can't migrate")
            from aerich import Command
            command = Command(tortoise_config=self._get_connection_setting(), app=self._app_name,
                              location=self.db_conf['migrations_path'])
            await self._migrations(schema_exists, command)
        if not schema_exists:
            await self.create_schema()
            await self.fill_db_data()

    async def close(self):
        await connections.close_all()

    async def create_schema(self, generate_schemas: bool = True):
        await self.conn.execute_script(f"CREATE SCHEMA IF NOT EXISTS {self.db_conf['db_schema']};")
        if generate_schemas:
            await Tortoise.generate_schemas()



from dependency_injector.wiring import Provide, inject
from sanic import Request

from web_foundation.workers.io.http.chaining import InputContext

@inject
async def protect_user(r: Request,app_container=Provide["<container>"]):
    print(app_container)
    return {}



from typing import List, Type

from sanic import json
from sanic.exceptions import NotFound
import json as py_json

from web_foundation.app.infrastructure.database.middleware import AccessDMW
from web_foundation.app.infrastructure.database.models import AbstractDbModel
from web_foundation.workers.io.http.chaining import InputContext


async def exec_access(inc: InputContext,
                      model: AbstractDbModel,
                      fetch_fields: List[str] = None,
                      middleware: Type[AccessDMW] = AccessDMW
                      ):
    """
    Access to models
    """
    match inc.request.method:
        case "GET":
            kwargs = inc.r_kwargs
            entity_id = kwargs.pop("entity_id", None)
            if entity_id is None:
                limit = inc.request.args.get("limit")
                offset = inc.request.args.get("offset")
                limit = int(limit) if limit and limit.isdigit() else 100
                offset = int(offset) if offset and offset.isdigit() else None
                # order_by = inc.request.args.get("order_by")  # for openapi. Variable forwarded to read_all in kwargs
                for ar, val in inc.request.args.items():
                    if ar in ["limit", "offset"]:
                        continue
                    val = val[0]
                    if val.lower() == 'false':
                        val = False
                    elif val.lower() == 'true':
                        val = True
                    elif val.isdigit():
                        val = int(val)
                    elif "." in val and val.replace('.', '').isdigit():
                        val = float(val)
                    elif val.startswith("["):
                        val = py_json.loads(val)
                    kwargs[ar] = val

                retrieved_all, total = await middleware.read_all(model, inc.identity, limit, offset,
                                                                fetch_fields=fetch_fields, **kwargs)
                res = [await model.values_dict() for model in retrieved_all]
                return res
            retrieved = await middleware.read(model, entity_id, inc.identity, fetch_fields=fetch_fields, **kwargs)
            if retrieved:
                result = await retrieved.values_dict()
                return result
            else:
                raise NotFound()

        case "POST":
            assert inc.dto
            retrieved = await middleware.create(model, inc.identity, inc.dto, **inc.r_kwargs)
            if isinstance(retrieved, list):
                return [await i.values_dict() for i in retrieved]
            return await retrieved.values_dict()

        case "PATCH":
            assert inc.dto
            entity_id = inc.r_kwargs.pop("entity_id")
            assert entity_id
            retrieved = await middleware.update(model, entity_id, inc.identity, inc.dto, **inc.r_kwargs)
            return await retrieved.values_dict()

        case "DELETE":
            entity_id = inc.r_kwargs.pop("entity_id")
            assert entity_id
            retrieved = await middleware.delete(model, entity_id, inc.identity, **inc.r_kwargs)
            return await retrieved.values_dict()



import asyncio
from pathlib import Path

import loguru
from dependency_injector import providers, containers
from pydantic import BaseSettings
from tortoise import fields

from web_foundation.app.app import AppContainer
from web_foundation.app.events.base import AeEvent
from web_foundation.app.infrastructure.database.models import AbstractDbModel
from web_foundation.app.resources.database import DatabaseResource, DbConfig
from web_foundation.app.service import Service
from web_foundation.kernel import GenericIMessage
from web_foundation.workers.io.http.chaining import chain
from web_foundation.workers.io.http.routers.dict_router import DictRouter
from web_foundation.workers.io.http.routes_dict import routes_dict
from web_foundation.workers.io.rt_connection import RtConnection
from web_foundation.workers.io.sse.connection import SseRtMessage
from web_foundation.workers.io.worker import ServerConfig, create_io_workers, subscribe_worker_rt_callback


class AppConf(BaseSettings):
    app_name: str
    server: ServerConfig
    database: DbConfig


class AnyModel(AbstractDbModel):
    some_data = fields.TextField()


class TicketService(Service):
    def __init__(self):
        super().__init__()


@containers.copy(AppContainer)
class MyContainer(AppContainer):
    ticket_service = providers.Singleton(TicketService)
    database = providers.Resource(DatabaseResource, app_name="app_name",
                                  db_config=AppContainer.app_config.database,
                                  modules=[__name__])


async def resolve_send_event(conn: RtConnection, msg: GenericIMessage) -> SseRtMessage | None:
    loguru.logger.warning(conn)
    loguru.logger.warning(msg)
    return SseRtMessage(event_id=str(msg.index), event_name="new_event", data={"event": "aaaaaaa"})


async def main():
    container = MyContainer()
    container.app_config.from_pydantic(
        AppConf(app_name="first_app", server=ServerConfig(host="0.0.0.0", port=8000),
                database=DbConfig(
                    host="localhost",
                    port="5432",
                    database="testesmdb",
                    user="testuser",
                    password="Fedora132",
                    db_schema="ae",
                    with_migrations=False,
                    migrations_path=Path("")
                )))
    # router = YamlRouter(Path("../applyed_files/config/routing.yaml"), chain)
    router = DictRouter(routes_dict, chain)
    workers = create_io_workers(container.app_config.server(), router, workers_num=2)
    subscribe_worker_rt_callback(*workers, event_type=AeEvent, callback=resolve_send_event, use_nested_classes=True)

    # exexitor = TaskExecutor()
    # exexitor.configure("exec_worker")
    # container.app().add_worker(exexitor)

    container.app().add_worker(workers)
    loguru.logger.warning(container.app().dispatcher.channels)
    loguru.logger.warning([{i.channel.worker_name: i.channel._listeners} for i in container.app().workers.values()])
    # htpp_srv = HttpServer(container.app_config.server())
    # htpp_srv.configure("name", yaml_router=router, sock=socket)
    # htpp_srv_2 = HttpServer(container.app_config.server())
    # htpp_srv_2.configure("name2", yaml_router=router, sock=socket)
    #
    # container.app().add_worker(htpp_srv_2)
    container.app().add_service(container.ticket_service())
    container.app().wire_app(container)

    # container.wire_this()
    await container.app().run()


if __name__ == '__main__':
    asyncio.run(main())



from enum import Enum
from ipaddress import IPv4Address
from typing import TypeVar, List, Any, Dict

from tortoise import fields
from tortoise.fields import ReverseRelation
from tortoise.models import Model
from tortoise.queryset import QuerySet

from .utils import string_from_db_date


class AbstractDbModel(Model):
    app_name: str = ""
    id = fields.IntField(pk=True)

    async def values_dict(self, m2m_fields: bool = False, fk_fields: bool = False, backward_fk_fields=False,
                          drop_cols: List[str] = None, iso_date_format=True, all_fetched=True) -> Dict[str, Any]:
        def _field_in_drop(field: str):
            if drop_cols and field in drop_cols:
                return True
            return False

        t_d = {}
        for k, v in self.__dict__.items():
            if _field_in_drop(k):
                continue
            v = string_from_db_date(v, iso_date_format)
            if isinstance(v, IPv4Address):
                v = str(v)
            if isinstance(v, Enum):
                v = v.value
            if not k.startswith('_'):
                t_d.update({k: v})
        if fk_fields or all_fetched:
            for field in self._meta.fk_fields:
                if _field_in_drop(field):
                    continue
                model = getattr(self, field)
                if isinstance(model, QuerySet):
                    if not fk_fields and all_fetched:
                        continue
                    model = await model
                if model:
                    t_d.update({field: await model.values_dict()})
        if m2m_fields or all_fetched:
            for field in self._meta.m2m_fields:
                if _field_in_drop(field):
                    continue
                models = getattr(self, field)
                if not models._fetched:
                    if not m2m_fields and all_fetched:
                        continue
                    models = await models
                t_d.update({field: [await i.values_dict() for i in models if i]})
        if backward_fk_fields:
            for field in self._meta.backward_fk_fields:
                if _field_in_drop(field):
                    continue
                model = getattr(self, field)
                if isinstance(model, ReverseRelation):
                    model = await model.all()
                if model:
                    t_d.update({field: [await i.values_dict() for i in model if i]})
        return t_d

    @classmethod
    def relate(cls, model_str: str):
        return f"{cls.app_name}.{model_str}"

    class Meta:
        abstract = True


GenericDbType = TypeVar("GenericDbType", bound=AbstractDbModel, contravariant=True)



from sanic.errorpages import exception_response
from sanic.handlers import ErrorHandler
from sanic.exceptions import SanicException

from wf_old.workers.web.erorrs.application import ApplicationError


class StatusCodes:
    APPLICATION_CODE = 409
    REPORT_CODE = 501


class ErrorHandlerExt(ErrorHandler):
    def default(self, request, exception):
        """Convert ApplicationError to SanicException"""
        if isinstance(exception, ApplicationError):
            exception = type(exception.__class__.__name__,
                             (SanicException,),
                             {"message": exception.message if exception.message else "",
                              "status_code": StatusCodes.APPLICATION_CODE})()
        else:
            self.log(request, exception)
        fallback = ErrorHandler._get_fallback_value(self, request.app.config)
        return exception_response(
            request,
            exception,
            debug=self.debug,
            base=self.base,
            fallback=fallback,
        )



from dataclasses import dataclass
from typing import List, TypeVar, Tuple, Generic

from pydantic import BaseModel as PdModel
from tortoise.exceptions import IntegrityError, FieldError
from tortoise.queryset import QuerySet, MODEL

from web_foundation.app.errors.application.application import InconsistencyError
from web_foundation.app.infrastructure.database.models import GenericDbType
from web_foundation.app.infrastructure.database.utils import integrity_error_format
from web_foundation.workers.io.http.chaining import ProtectIdentity

EntityId = TypeVar("EntityId", bound=int)


@dataclass
class AccessProtectIdentity(Generic[GenericDbType], ProtectIdentity):
    pass


class DatabaseMiddleware:
    pass


class RelateDMW(DatabaseMiddleware):
    @staticmethod
    async def fetch_related(model: GenericDbType, query: QuerySet[MODEL],
                            fetch_fields: List[str] | None = None) -> QuerySet[MODEL]:
        if fetch_fields:
            for_select = {field for fk_field in model._meta.fk_fields for field in fetch_fields if
                          fk_field == field}
            for_prefetch = set(fetch_fields) - for_select
            if for_select:
                query = query.select_related(*for_select)
            if for_prefetch:
                query = query.prefetch_related(*for_prefetch)
        return query


class AccessDMW(DatabaseMiddleware):
    @staticmethod
    async def get_entity(model: GenericDbType,
                         entity_id: EntityId) -> GenericDbType:  # TODO TYPED USER
        entity = await model.get_or_none(id=entity_id)
        if not entity:
            raise InconsistencyError(message=f"{model.__name__} not found")  # type: ignore
        return entity

    @staticmethod
    async def create(model: GenericDbType, protection: AccessProtectIdentity, dto: PdModel,
                     **kwargs) -> GenericDbType:
        entity_kwargs = {field: value for field, value in dto.dict().items() if value is not None}
        try:
            entity = await model.create(**entity_kwargs)
        except IntegrityError as exception:
            raise integrity_error_format(exception)
        except ValueError as exception:
            raise InconsistencyError(exception)
        return entity

    @staticmethod
    async def read(model: GenericDbType, entity_id: EntityId, protection: AccessProtectIdentity,
                   fetch_fields: List[str] | None = None,
                   **kwargs) -> GenericDbType | None:
        query = model.filter(id=entity_id)
        if kwargs:
            query = query.filter(**kwargs)
        query = await RelateDMW.fetch_related(model, query, fetch_fields)
        entity = await query.first()
        return entity

    @staticmethod
    async def read_all(model: GenericDbType, protection: AccessProtectIdentity, limit: int = None, offset: int = None,
                       fetch_fields: List[str] = None,
                       **kwargs) -> Tuple[List[GenericDbType], int]:
        query = model.filter()
        if kwargs:
            if "order_by" in kwargs:
                query = query.order_by(kwargs.pop("order_by"))
            query = query.filter(**kwargs)

        if limit:
            query = query.limit(limit)
        if offset:
            query = query.offset(offset)
        query = await RelateDMW.fetch_related(model, query, fetch_fields)
        try:
            entities = await query
            query._limit = None
            query._offset = None
            total = await query.count()
        except FieldError:
            raise InconsistencyError(message="Incorrect filter")
        return entities, total

    @staticmethod
    async def update(model: GenericDbType, entity_id: EntityId, protection: AccessProtectIdentity, dto: PdModel,
                     **kwargs) -> GenericDbType:
        entity = await AccessDMW.get_entity(model, entity_id)
        entity_kwargs = {field: value for field, value in dto.dict().items() if value is not None}
        for field, value in entity_kwargs.items():
            setattr(entity, field, value)
        try:
            await entity.save(update_fields=list(entity_kwargs.keys()))
        except IntegrityError as exception:
            raise integrity_error_format(exception)
        except ValueError as exception:
            raise InconsistencyError(exception)
        return entity

    @staticmethod
    async def delete(model: GenericDbType, entity_id: EntityId, protection: AccessProtectIdentity,
                     **kwargs) -> GenericDbType:
        assert entity_id is not None
        entity = await AccessDMW.get_entity(model, entity_id)
        await entity.delete()
        return entity



import inspect
import re
from typing import Dict, Callable, Type
from pydantic import BaseModel as PdModel

import sanic_ext
from sanic import Sanic
from sanic_ext.extensions.openapi.builders import SpecificationBuilder, OperationStore

from web_foundation.workers.io.http.routers.ext_router import ExtRouter, RouteMethodConf, RouteConf


def get_definition_with_correct_name(definitions: Dict) -> Dict:
    _definitions = {}
    for defin, _schema in definitions.copy().items():
        if "." in defin:
            _definitions[defin.split(".")[-2]] = _schema
        else:
            _definitions[defin] = _schema
    return _definitions


def get_model_schema(model: Type[PdModel]):
    """ Create schema from pydantic model and add all references to OpenAPI schemas """
    schema = model.schema(ref_template="#/components/schemas/{model}")
    components = {}
    if "definitions" in schema:
        definitions = get_definition_with_correct_name(schema.pop("definitions"))
        components.update(definitions)
    components.update({model.__name__: schema})
    spec = SpecificationBuilder()
    for component_name, component in components.items():
        spec._components["schemas"].update({component_name: component})
    return schema


def add_openapi_spec(uri: str, method_name: str, func: Callable, handler: Callable,
                     in_dto: Type[PdModel], out_dto: Type[PdModel]):
    # --- add query args ---#
    func_text = inspect.getsource(func)
    func_query_args = re.findall(r"request.args.get\(\"(\w*)\"\)", func_text)
    if func_query_args:
        for func_arg in func_query_args:
            handler = sanic_ext.openapi.parameter(func_arg)(handler)
    # --- add request body ---#
    if in_dto:
        if issubclass(in_dto, PdModel):
            schema = get_model_schema(in_dto)
            OperationStore()[handler].requestBody = {"content": {"application/json": {'schema': schema}}}
    # --- add response ---#
    if out_dto:
        if issubclass(out_dto, PdModel):
            schema = get_model_schema(out_dto)
            OperationStore()[handler].responses[200] = {"content": {"application/json": {'schema': schema}}}
    # --- add tag ---#
    handler = sanic_ext.openapi.tag(uri.split("/")[1].capitalize())(handler)  # first path word

    # --- decription --- #
    # f = sanic_ext.openapi.description("hujkiikasdf;awerfl;jkasdfklj;asdfjkl;")(handler)
    # f = sanic_ext.openapi.summary("l;kasdjllww")(handler)

    # --- set operation id --- #
    handler = sanic_ext.openapi.operation(f"{method_name}~{uri}")(handler)
    return handler


class DictRouter(ExtRouter):
    _router_conf: Dict
    chaining: Callable
    versioning: bool

    def __init__(self, routes_config: Dict, chaining: Callable, versioning: bool = True):
        super().__init__()
        self.chaining = chaining
        self._router_conf = routes_config
        self.chains = []
        self.versioning = versioning
        self._parse()

    def _parse(self):
        for app_route in self._router_conf.get("apps"):
            version_prefix = app_route.get("version_prefix")
            version_prefix = version_prefix if version_prefix else "/api/v"
            for endpoint, versions in app_route.get("endpoints").items():
                for version, params in versions.items():
                    methods_confs = []
                    endpoint_handler = params.pop("handler", None)
                    endpoint_protector = params.pop("protector", None)
                    endpoint_response_fabric = params.pop("response_fabric", None)
                    for method_name, method_params in params.items():
                        target_func = method_params.get('handler')
                        target_func = target_func if target_func else endpoint_handler
                        protector = method_params.get("protector")
                        protector = protector if protector else endpoint_protector
                        in_dto = method_params.get("in_dto")
                        out_dto = method_params.get("out_dto")
                        response_fabric = method_params.get("response_fabric")
                        response_fabric = response_fabric if response_fabric else endpoint_response_fabric
                        if response_fabric:
                            chain = self.chaining(
                                protector=protector,
                                in_struct=in_dto,
                                response_fabric=response_fabric)(target_func)
                        else:
                            chain = self.chaining(
                                protector=protector,
                                in_struct=in_dto)(target_func)

                        chain = add_openapi_spec(uri=endpoint, method_name=method_name, func=target_func,
                                                 handler=chain, in_dto=in_dto, out_dto=out_dto)

                        methods_confs.append(RouteMethodConf(method_name=method_name,
                                                             protector=protector,
                                                             in_dto=in_dto,
                                                             out_dto=out_dto,
                                                             handler=target_func,
                                                             chain=chain,
                                                             version_prefix=version_prefix,
                                                             version=version
                                                             ))

                    route = RouteConf(app_name=app_route.get("app_name"),
                                      path=endpoint,
                                      methods=methods_confs)
                    self.chains.append(route)

    def apply_routes(self, app: Sanic, **kwargs):
        """
        If you want use chaining, please use partial(chain,validation_fnc=...,response_fabric=...)
        :param **kwargs:
        :param chaining:
        :return:
        """
        for route_conf in self.chains:
            for method_conf in route_conf.methods:
                if self.versioning:
                    app.add_route(uri=route_conf.path, methods={method_conf.method_name.upper()},
                                  handler=method_conf.chain,
                                  version_prefix=method_conf.version_prefix, version=method_conf.version)
                else:
                    app.add_route(uri=route_conf.path, methods={method_conf.method_name.upper()},
                                  handler=method_conf.chain)

        # warn(f"Can't find dto or handler in imported module, suppressed exception: {e.__str__()}")



from __future__ import annotations

from typing import TypeVar

from dependency_injector import containers
from dependency_injector.wiring import inject, Provide

from web_foundation.kernel import Isolate


class Worker(Isolate):
    _configured: bool

    def __init__(self, *args, **kwargs):
        self._configured = False

    @property
    def configured(self) -> bool:
        return self._configured

    @inject
    async def _init_resources(self, app=None, app_container=Provide["<container>"]):
        await app_container.init_resources()
        for service in app_container.app().services.values():
            service.worker = self
        pass

    def configure(self, name: str, debug=True, *args, **kwargs):
        self._configure_isolate(name, debug)
        self._configure(*args, **kwargs)
        self._configured = True

    def _configure(self, *args, **kwargs):
        pass

    def wire_worker(self, container: containers.DeclarativeContainer) -> None:
        raise NotImplementedError

    async def _run(self):
        raise NotImplementedError

    def _close(self):
        pass


GenWorker = TypeVar("GenWorker", bound=Worker)



from dependency_injector.wiring import Provide, inject
from sanic import Request


async def some_handler(request):
    return {}


@inject
async def some_protect(request: Request, app_container=Provide["<container>"]):
    print(app_container)
    return None


class SomeDto:
    def __init__(self, *asfasgasg, **kwargs):
        pass


class SomeDto2:
    def __init__(self, *asfasgasg, **kwargs):
        pass


routes_dict = {
    "apps": [
        {
            "app_name": "ae_app",
            "version_prefix": "/api/v",
            "endpoints": {
                "/ticket": {
                    "v14": {
                        "handler": some_handler,
                        "protector": some_protect,
                        "get": {},
                        "post": {
                            "handler": some_handler,
                            "protector": some_protect,
                        }
                    }
                }
            }
        }
    ]
}



import asyncio
import datetime
import uuid
from asyncio import Task
from dataclasses import dataclass, field
from functools import partial
from typing import Dict, Any
from time import time_ns
from dependency_injector import containers
from loguru import logger
from inspect import iscoroutine
from web_foundation.workers.background.executor.message import TaskIMessage
from web_foundation.workers.worker import Worker


@dataclass
class BackgroundTask:
    task: Task[Any]
    status: str
    id: uuid.UUID = field(default_factory=uuid.uuid4)
    exec_time: int = field(default=0)
    done_time: int = field(default=0)
    call_time: float = field(default=0)
    finish_time: datetime.datetime | None = field(default_factory=datetime.datetime.now)
    start_time: datetime.datetime | None = field(default_factory=datetime.datetime.now)
    error: BaseException | None = field(default=None)


class TaskExecutor(Worker):
    _runned_task: Dict[uuid.UUID, BackgroundTask]
    _completed_tasks: Dict[uuid.UUID, BackgroundTask]
    _errors_task: Dict[uuid.UUID, BackgroundTask]

    def wire_worker(self, container: containers.DeclarativeContainer) -> None:
        pass

    def __init__(self, **kwargs):
        self._runned_task = {}
        self._completed_tasks = {}
        self._errors_task = {}
        super().__init__(**kwargs)

    async def _run(self):
        self.channel.add_event_listener(TaskIMessage, self.task_handler)
        await self.channel.listen_consume()

    async def task_handler(self, message: TaskIMessage):
        task = asyncio.create_task(message.task(*message.args, **message.kwargs))  # type: ignore
        arc_task = BackgroundTask(task=task, status="pending")
        arc_task.exec_time = time_ns()
        self._runned_task[arc_task.id] = arc_task
        # TODO Think about responses
        task.add_done_callback(partial(self.on_task_done, arc_task=arc_task))

    def on_task_done(self, task: Task, arc_task: BackgroundTask):
        arc_task.done_time = time_ns()
        arc_task.call_time = (arc_task.done_time - arc_task.exec_time) / 1000000000
        exeption = task.exception()
        if exeption:
            self._errors_task[arc_task.id] = self._runned_task.pop(arc_task.id)
            arc_task.status = f"error {exeption.__str__()}"
            arc_task.error = exeption
            if self.debug:
                logger.warning(f"Task {arc_task.id} provide error {exeption.__str__()} in {arc_task.call_time}")
        else:
            self._completed_tasks[arc_task.id] = self._runned_task.pop(arc_task.id)
            if self.debug:
                logger.info(f"Task {arc_task.id} done in {arc_task.call_time}")
            arc_task.status = "success"

    def _close(self, *args, **kwargs):
        pass

    def get_task_stats(self):
        result = {}
        for dictionary in [self._runned_task, self._errors_task, self._completed_tasks]:
            result.update(dictionary)
        return result



from dataclasses import dataclass
from typing import List, Type

from pydantic import BaseModel
from sanic import Sanic
from sanic.router import Router

from web_foundation.workers.io.http.chaining import HandlerType, Protector


@dataclass
class RouteMethodConf:
    method_name: str
    protector: Protector | None
    in_dto: Type[BaseModel] | None
    out_dto: Type[BaseModel] | None
    handler: HandlerType
    chain: HandlerType
    version_prefix: str | None
    version: str | None


@dataclass
class RouteConf:
    app_name: str
    path: str
    methods: List[RouteMethodConf]


class ExtRouter(Router):
    chains: List[RouteConf]

    def apply_routes(self, app: Sanic, *args, **kwargs):
        raise NotImplementedError



from typing import Union, Dict

import loguru


class ApplicationError(Exception):
    error_type: str = "C"
    default_class: int = 999
    default_subclass: int = 999
    context: Dict[str, Union[str, int]] = None

    def __init__(self, ex=None, message=None, context=None):
        if ex:
            self.message = str(ex)
        else:
            self.message = message
        if context:
            self.context = context
        elif not self.context:
            self.set_default_context()

        super().__init__(self.message)

    def set_default_context(self):
        self.context = {
            "type": self.error_type,
            "code": f"{self.error_type}-{self.default_class}-{self.default_subclass}",
            "class": self.default_class,
            "subclass": self.default_subclass,
            "comment": self.message
        }


class InconsistencyError(ApplicationError):
    error_type: str = "L"

    def __init__(self, ex=None, message=None, context=None):
        super().__init__(ex=ex, message=message, context=context)


class ValidationError(ApplicationError):
    error_type: str = "V"


class IntegrationError(ApplicationError):
    error_type: str = "I"



from typing import Type

from sanic import Request
from pydantic import ValidationError as PdValidationError, BaseModel as PdModel

from web_foundation.app.errors.application.application import ValidationError


def validate_dto(dto_cls: Type[PdModel] | None,request: Request) -> PdModel | None:
    if not dto_cls:
        return None
    try:
        dto = dto_cls(**request.json)
        return dto
    except PdValidationError as ex:
        failed_fields = ex.errors()
        fields = [field["loc"][-1] for field in failed_fields]
        commment_str = "Some of essential params failed : " + ", ".join(
            [field["loc"][-1] + " - " + field["msg"] for field in failed_fields])
        message = commment_str
        context = {
            "fields": fields,
            "comment": commment_str
        }
        raise ValidationError(message=message,
                              context=context)


def dto_validation_error_format(exeption: PdValidationError):
    failed_fields = exeption.errors()
    commment_str = "Some of essential params failed : " + ", ".join(
        [str(field["loc"][-1]) + " - " + field["msg"] for field in failed_fields])
    return commment_str



from dataclasses import dataclass
from functools import wraps
from typing import Any, Callable, Type, Coroutine, TypeVar, Generic
from typing import Union, Dict, List

from pydantic import BaseModel as PdModel
from sanic import Request, json, HTTPResponse

from web_foundation.utils.validation import validate_dto

DtoType = TypeVar("DtoType", bound=PdModel)


@dataclass
class ProtectIdentity():
    pass


ProtectIdentityType = TypeVar("ProtectIdentityType", bound=ProtectIdentity)


@dataclass
class InputContext(Generic[DtoType, ProtectIdentityType]):
    request: Request
    dto: DtoType | None
    identity: ProtectIdentity | None
    r_kwargs: Dict


Protector = Callable[[Request], Coroutine[Any, Any, ProtectIdentityType | None]]
DtoValidator = Callable[[Type[DtoType], Request], PdModel]
HandlerType = Callable[[InputContext], Coroutine[Any, Any, HTTPResponse]]

JSON = Union[Dict[str, Any], List[Any], int, str, float, bool, Type[None]]
TypeJSON = Union[Dict[str, 'JSON'], List['JSON'], int, str, float, bool, Type[None]]


async def protect(request: Request) -> ProtectIdentity | None:
    return None


def chain(protector: Protector = protect,
          in_struct: Type[PdModel] | None = None,
          validation_fnc: DtoValidator = validate_dto,
          response_fabric: Callable[[TypeJSON], HTTPResponse] = json):
    def called_method(target: HandlerType):
        @wraps(target)
        async def f(*args, **kwargs):
            req: Request = args[0]
            prot_identity = await protector(req) if protector else None
            if in_struct:
                validated = validation_fnc(in_struct, req)
            else:
                validated = None
            incoming = InputContext(req, validated, prot_identity, kwargs)
            ret_val = await target(incoming)
            return response_fabric(ret_val)

        return f

    return called_method


async def any_protect():
    pass


class Str(PdModel):
    name: str


async def handler(inc: InputContext):
    return {}



from functools import partial
from typing import Callable, Any, Coroutine, TypeVar, Generic

from web_foundation.kernel import IMessage

BgTask = TypeVar("BgTask", Callable[..., Coroutine[Any, Any, Any]], partial)


class TaskIMessage(Generic[BgTask], IMessage):
    message_type = "background_task"
    task: BgTask
    args: Any | None
    kwargs: Any | None

    def __init__(self, task: BgTask, args: Any = None, kwargs: Any | None = None):
        super().__init__()
        self.task = task
        self.args = args
        self.kwargs = kwargs



import loguru
from sanic.exceptions import SanicException
from sanic.exceptions import Unauthorized as SanicUnauthorized

from web_foundation.app.errors.application.application import ApplicationError


class BaseAuthException(SanicException, ApplicationError):
    error_type: str = "L"
    default_class: int = 1
    default_subclass: int = 999


class InvalidToken(BaseAuthException):
    pass


class AuthenticationFailed(BaseAuthException):
    status_code = 401

    def __init__(self, message="Authentication failed.", **kwargs):
        super().__init__(message, **kwargs)


class ScopesFailed(BaseAuthException):
    status_code = 401

    def __init__(self, message="Authentication failed.", **kwargs):
        super().__init__(message, **kwargs)


class MissingAuthorizationHeader(BaseAuthException):
    status_code = 400

    def __init__(self, message="Authorization header not present.", **kwargs):
        super().__init__(message, **kwargs)


class MissingAuthorizationCookie(BaseAuthException):
    status_code = 400

    def __init__(self, message="Authorization cookie not present.", **kwargs):
        super().__init__(message, **kwargs)


class MissingAuthorizationQueryArg(BaseAuthException):
    status_code = 400

    def __init__(
            self, message="Authorization query argument not present.", **kwargs
    ):
        super().__init__(message, **kwargs)


class InvalidAuthorizationHeader(BaseAuthException):
    status_code = 400

    def __init__(self, message="Authorization header is invalid.", **kwargs):
        super().__init__(message, **kwargs)


class InvalidCustomClaim(BaseAuthException):
    status_code = 500

    def __init__(self, message="Custom claim is invalid.", **kwargs):
        super().__init__(message, **kwargs)


class InvalidCustomClaimError(BaseAuthException):
    status_code = 401

    def __init__(self, message="Custom claim value was not met.", **kwargs):
        super().__init__(message, **kwargs)


class InvalidVerification(BaseAuthException):
    status_code = 500

    def __init__(
            self,
            message="Verifications must be a callable object "
                    "returning a boolean value.",
            **kwargs
    ):
        super().__init__(message, **kwargs)


class InvalidVerificationError(BaseAuthException):
    status_code = 401

    def __init__(self, message="Verifications were not met.", **kwargs):
        super().__init__(message, **kwargs)


class AuthenticateNotImplemented(BaseAuthException):
    status_code = 500

    def __init__(
            self,
            message="Sanic JWT initialized without providing an authenticate "
                    "method.",
            **kwargs
    ):
        super().__init__(message, **kwargs)


class RefreshTokenNotImplemented(BaseAuthException):
    status_code = 500

    def __init__(
            self, message="Refresh tokens have not been enabled.", **kwargs
    ):
        super().__init__(message, **kwargs)


class ScopesNotImplemented(BaseAuthException):
    status_code = 500

    def __init__(
            self,
            message="Scopes have not been enabled. Initialize with "
                    "add_scopes_to_payload to provide scoping.",
            **kwargs
    ):
        super().__init__(message, **kwargs)


class UserSecretNotImplemented(BaseAuthException):
    status_code = 500

    def __init__(
            self, message="User secrets have not been enabled.", **kwargs
    ):
        super().__init__(message, **kwargs)


class MissingRegisteredClaim(BaseAuthException):
    status_code = 500

    def __init__(
            self,
            message="One or more claims have been registered, but your "
                    "extend_payload() method does not supply them. ",
            missing=None,
            **kwargs
    ):
        if missing:  # noqa
            message += str(missing)
        super().__init__(message, **kwargs)


class MeEndpointNotSetup(BaseAuthException):
    status_code = 500

    def __init__(
            self,
            message="/me endpoint has not been setup. Pass retrieve_user if "
                    "you with to proceeed.",
            **kwargs
    ):
        super().__init__(message, **kwargs)


class InvalidRetrieveUserObject(BaseAuthException):
    status_code = 500

    def __init__(
            self,
            message="The retrieve_user method should return either a dict or "
                    "an object with a to_dict or __json__ method.",
            **kwargs
    ):
        super().__init__(message, **kwargs)


class InitializationFailure(BaseAuthException):
    status_code = 500

    def __init__(
            self,
            message="Sanic JWT was not initialized properly. It must be "
                    "instantiated on a sanic.Sanic or sanic.Blueprint "
                    "instance.",
            **kwargs
    ):
        super().__init__(message, **kwargs)


class Unauthorized(BaseAuthException, SanicUnauthorized):
    def __init__(self, message="Auth required.", **kwargs):
        super().__init__(message, scheme="Bearer", **kwargs)


class InvalidClassViewsFormat(BaseAuthException):
    def __init__(
            self,
            message="class_views should follow this format ('<SOME ROUTE>', "
                    "ClassInheritedFromBaseEndpoint)",
            **kwargs
    ):
        super().__init__(message, **kwargs)


class InvalidConfiguration(BaseAuthException):
    def __init__(self, message="", **kwargs):
        message = (
                "An invalid setting was passed to the Sanic JWT "
                "configuration: " + str(message)
        )
        super().__init__(message, **kwargs)


class InvalidPayload(BaseAuthException):
    status_code = 500

    def __init__(self, message="", **kwargs):
        message = (
            "Payload must be a dictionary with a key mapped to "
            "SANIC_JWT_USER_ID"
        )
        super().__init__(message, **kwargs)


class RequiredKeysNotFound(BaseAuthException):
    def __init__(
            self,
            message="You must provide both (valid) SANIC_JWT_PUBLIC_KEY and "
                    "SANIC_JWT_PRIVATE_KEY when using asymmetric "
                    "cryptographic algorithms like RS*, EC* or PS*",
            **kwargs
    ):
        super().__init__(message, **kwargs)


class ProvidedPathNotFound(BaseAuthException):
    def __init__(
            self, message="The Path object given is not a valid file", **kwargs
    ):
        super().__init__(message, **kwargs)


class LoopNotRunning(BaseAuthException):
    def __init__(
            self, message="The asyncio loop is not currently running", **kwargs
    ):
        super().__init__(message, **kwargs)



from __future__ import annotations

import asyncio
from asyncio import Task
from typing import Generic, Any, Callable, Coroutine

import loguru
from loguru import logger

from web_foundation.kernel import GenericIMessage
from web_foundation.workers.io.http.chaining import InputContext


class RtMessage:

    def _prepare(self, *args, **kwargs) -> str | bytes:
        raise NotImplementedError

    @property
    def to_sent(self) -> str | bytes:
        return self._prepare()

    @classmethod
    def ping_message(cls):
        raise NotImplementedError


class WriteableObj:
    obj: Any

    def __init__(self, obj: Any):
        self.obj = obj

    async def write(self, message: Any) -> None:
        pass


class RtConnection(Generic[GenericIMessage]):
    debug: bool
    ping_timeout: float
    listen_timeout: float
    writeable: WriteableObj
    input_ctx: InputContext
    _last_event_id: int
    _ping_task: Task
    _ping_enable: bool
    _on_disconnect: Callable[[], Coroutine]

    def __init__(self, input_ctx: InputContext, writeable: WriteableObj, debug: bool = True, ping_enable: bool = True,
                 ping_timeout=None,
                 listen_timeout=None):
        self.writeable = writeable
        self.input_ctx = input_ctx
        self.debug = debug
        self._ping_enable = ping_enable
        self.ping_timeout = ping_timeout if ping_timeout else 5
        self.listen_timeout = listen_timeout if listen_timeout else 0.1
        self._last_event_id = 0

    def _ping_msg(self) -> str | bytes:
        raise NotImplementedError

    async def _ping(self):
        while True:
            await self.writeable.write(self._ping_msg())
            if self.debug:
                logger.debug(f"Sent ping to {self.input_ctx.request.ip}")
            await asyncio.sleep(self.ping_timeout)

    async def freeze_request(self, on_disconnect: Callable[[], Coroutine]):
        self._on_disconnect = on_disconnect
        await self._ping()

    async def send_after_call(self, event: GenericIMessage, resolve_callback: RtEventCallback):
        msg = await resolve_callback(self, event)
        if msg:
            await self.writeable.write(msg.to_sent)

    async def __aenter__(self):
        pass

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        loguru.logger.warning("as[gjasbngkasngjansg")
        await self._on_disconnect()
        if self.debug:
            logger.debug(
                f"RtConnection({self.__class__.__name__}) on {self.input_ctx.request.ip} disconnected, ping task canceled")

    def _close_condition(self) -> bool:
        raise NotImplementedError

    @classmethod
    async def accept_connection(cls, input_ctx: InputContext,
                                ping_enable: bool = True,
                                ping_timeout=None,
                                listen_timeout=None,
                                debug: bool = True) -> RtConnection:
        raise NotImplementedError

    @classmethod
    async def _construct_writable(cls, input_ctx: InputContext, *args, **kwargs) -> WriteableObj:
        raise NotImplementedError


RtEventCallback = Callable[[RtConnection, GenericIMessage], Coroutine[Any, Any, RtMessage | None]]

#
# async def stream_queued_events(self, resolve_callback: ResolvedCallback) -> None:
#
#     self._response = await self.input_ctx.request.respond(content_type="text/event-stream")
#     ping_task = self.input_ctx.request.app.loop.create_task(_ping(self.ping_timeout))
#     while True:
#         if not self._response.stream:
#             ping_task.cancel()
#             print(f"Sse on {self.input_ctx.request.ip} disconnected, ping task canceled")
#             break
#         if not self.connection_queue.empty():
#             resolved = await resolve_callback(self, await self.connection_queue.get())
#             if resolved.disconnect:
#                 break
#             if resolved.event_data and resolved.event_name != "":
#                 ev_payload = SseProto(event_id=str(self._last_event_id), event_name=resolved.event_name,
#                                       data=resolved.event_data).to_send
#                 await self._response.send(ev_payload.encode())
#                 self._last_event_id += 1
#         await asyncio.sleep(self.listen_timeout)



import asyncio
from abc import ABCMeta, abstractmethod
from asyncio import Future

from aioprocessing import AioProcess
from web_foundation.kernel.channel import IChannel


class Isolate(metaclass=ABCMeta):
    debug: bool
    _name: str
    _channel: IChannel
    _proc: AioProcess
    _created = False

    @property
    def name(self):
        return self._name

    @property
    def created(self):
        return self._created

    def _configure_isolate(self, name: str, debug=True):
        self.debug = debug
        self._name = name
        self._proc = AioProcess(target=self._startup)
        self._channel = IChannel(self._name, debug=debug)
        self._created = True

    @abstractmethod
    async def _run(self):
        pass

    @abstractmethod
    def _close(self, *args, **kwargs):
        pass

    def close(self, *args, **kwargs):
        self._close(*args, **kwargs)

    def _startup(self):
        asyncio.run(self._run())

    async def _exec(self):
        self._proc.start()

    def perform(self) -> Future:
        if not self._created:
            raise AttributeError("Call configure_isolate() before start perform")
        return asyncio.ensure_future(self._exec())

    @property
    def channel(self) -> IChannel:
        return self._channel

    @property
    def pid(self) -> int:
        return self._proc.pid

    @property
    def process(self) -> AioProcess:
        return self._proc

    def __str__(self):
        return f"{self.__class__.__name__}({self.name}) on pid: {self.pid}"



# import pyyaml module
import traceback
from dataclasses import dataclass
from functools import partial
from importlib import import_module
from pathlib import Path
from typing import List, Type, Dict, Callable
from warnings import warn

import yaml
from pydantic import BaseModel
from sanic import Sanic
from yaml.loader import SafeLoader

from web_foundation.app.errors.io.routing import YamlRouterLoadFileError
from web_foundation.workers.io.http.chaining import Protector, HandlerType
from web_foundation.workers.io.http.routers.ext_router import ExtRouter


@dataclass
class RouteMethodConf:
    method_name: str
    protector_path: str | None
    in_dto: str | None
    out_dto: str | None


@dataclass
class RouteConf:
    app_name: str
    scope: str
    path: str
    methods: List[RouteMethodConf]
    handler_path: str

    def dict_(self):
        return {"app_name": self.app_name, "scope": self.scope, "path": self.path}


@dataclass
class ImportedRouteMethodConf:
    method_name: str
    protector: Protector | None
    in_dto: Type[BaseModel] | None
    out_dto: Type[BaseModel] | None


@dataclass
class ImportedRouteConf:
    app_name: str
    scope: str
    path: str
    methods: List[ImportedRouteMethodConf]
    handler: HandlerType


class YamlRouter(ExtRouter):
    routes_confs: List[RouteConf]
    _routes_config: Path
    _route_protectors: Dict[str, Dict[str, str]]
    protectors: Dict[str, partial[Protector]]
    chains: List[ImportedRouteConf]
    chaining: Callable

    def __init__(self, routes_config: Path, chaining: Callable):
        super().__init__()
        self.chaining = chaining  # type: ignore
        self._routes_config = routes_config
        self._route_protectors = {}
        self.protectors = {}
        self.chains = []
        self.routes_confs = self._parce_file()

    def _parce_file(self) -> List[RouteConf]:
        try:
            with open(self._routes_config, 'r') as conf_f:
                yml = yaml.load(conf_f, Loader=SafeLoader)
                for app_name, data in yml["apps"].items():
                    if data["routes"].get("protectors"):
                        for protector_name, pr_settings in data["routes"].get("protectors").items():
                            pr_settings.update({"app_name": app_name})
                            self._route_protectors.update({protector_name: pr_settings})
                    endpoints = data["routes"]["endpoints"]
                    routes = []
                    for scope, params in endpoints.items():
                        base_protector = params.get("protector")
                        methods = []
                        for method, cfg in params.get("methods").items():
                            protector = cfg.get("protector")
                            if protector is not None:
                                protector = base_protector
                            methods.append(
                                RouteMethodConf(method_name=method, protector_path=protector,
                                                in_dto=cfg.get("in-dto"), out_dto=cfg.get("out-dto"))
                            )
                        route_cfg = RouteConf(app_name=app_name,
                                              scope=scope,
                                              path=params["path"],
                                              handler_path=params["handler"],
                                              methods=methods
                                              )
                        routes.append(route_cfg)
                    for pr_name, pr_settings in self._route_protectors.items():
                        app_name = pr_settings.pop("app_name")
                        fnc_name = pr_settings.pop("fnc")
                        if not fnc_name:
                            continue
                        splits_name = fnc_name.split(".")
                        module_name = ".".join([app_name, *splits_name[:-1]])
                        fnc = getattr(import_module(module_name), splits_name[-1])
                        self.protectors.update({pr_name: partial(fnc, **pr_settings)})
                return routes
        except Exception as e:
            raise YamlRouterLoadFileError(e)

    def _get_importable(self, app_name: str, scope: str, kind: str, path: str):
        if not path:
            return None
        module = import_module(".".join([app_name, "api", scope, "handlers"]))
        return getattr(module, path)

    def apply_routes(self, app: Sanic):
        """
        If you want use chaining, please use partial(chain,validation_fnc=...,response_fabric=...)
        :param chaining:
        :return:
        """
        try:
            for route in self.routes_confs:
                importer = partial(self._get_importable, app_name=route.app_name, scope=route.scope)
                imported_conf = ImportedRouteConf(handler=importer(path=route.handler_path, kind="handlers"),
                                                  **route.dict_(), methods=[])
                for method in route.methods:
                    imported_method = ImportedRouteMethodConf(method_name=method.method_name,
                                                              protector=self.protectors.get(method.protector_path),
                                                              in_dto=importer(path=method.in_dto, kind="dtos"),
                                                              out_dto=importer(path=method.out_dto, kind="dtos"),
                                                              )
                    imported_conf.methods.append(imported_method)
                    handler = self.chaining(
                        protector=imported_method.protector,
                        in_struct=imported_method.in_dto)(imported_conf.handler)
                    app.add_route(uri=route.path, methods={method.method_name.upper()}, handler=handler)
                self.chains.append(imported_conf)
        except ImportError as e:
            warn(f"Can't import from config path, suppressed exception: {e.__str__()}")
        except AttributeError as e:
            traceback.print_tb(e.__traceback__)
            warn(f"Can't find dto or handler in imported module, suppressed exception: {e.__str__()}")

#
# r = set_to_pipline(acc.uri)(wraps(target)(partial(target)))



import asyncio
from asyncio import Task, Future
from typing import Dict, List, Any

import loguru

from web_foundation.kernel import IChannel, GenericIMessage
from web_foundation.workers.worker import GenWorker


class IDispatcher:
    debug: bool = True
    channels: Dict[str, IChannel]
    _msg_global_index: int

    def __init__(self):
        self._msg_global_index = 0
        self.channels = {}

    def reg_worker(self, worker: GenWorker):
        self.channels.update({worker.name: worker.channel})

    async def on_channel_sent(self, msg: GenericIMessage):
        self._msg_global_index += 1
        msg.index = self._msg_global_index
        asyncio.create_task(self.broadcast(msg))

    async def broadcast(self, msg: GenericIMessage):
        for ch in self.channels.values():
            if self.debug:
                loguru.logger.debug(f"Sent {msg} to {ch.worker_name}")
            await ch.consume_pipe.write(msg)

    def preform(self) -> List[Future[Any]]:
        tasks = []
        for channel in self.channels.values():
            tasks.append(asyncio.ensure_future(
                channel.listen_produce(
                    self.on_channel_sent
                )))
        return tasks



from typing import Iterable

from collections import defaultdict
from typing import Dict, FrozenSet

import loguru
from sanic import Sanic, response
from sanic.router import Route


def _compile_routes_needing_options(
        routes: Dict[str, Route]
) -> Dict[str, FrozenSet]:
    needs_options = defaultdict(list)
    # This is 21.12 and later. You will need to change this for older versions.
    for route in routes.values():
        if "OPTIONS" not in route.methods:
            needs_options[route.uri].extend(route.methods)

    return {
        uri: frozenset(methods) for uri, methods in dict(needs_options).items()
    }


def _options_wrapper(handler, methods):
    def wrapped_handler(request, *args, **kwargs):
        nonlocal methods
        return handler(request, methods)

    return wrapped_handler


async def options_handler(request, methods) -> response.HTTPResponse:
    resp = response.empty()
    _add_cors_headers(resp, methods)
    return resp


def setup_options(app: Sanic):
    app.router.reset()
    needs_options = _compile_routes_needing_options(app.router.routes_all)
    for uri, methods in needs_options.items():
        try:
            app.add_route(
                _options_wrapper(options_handler, methods),
                uri,
                methods=["OPTIONS"],
            )
        except:
            pass
    app.router.finalize()


def _add_cors_headers(response, methods: Iterable[str]) -> None:
    allow_methods = list(set(methods))
    if "OPTIONS" not in allow_methods:
        allow_methods.append("OPTIONS")
    headers = {
        "Access-Control-Allow-Methods": ",".join(allow_methods),
        "Access-Control-Allow-Origin": "*",
        "Access-Control-Allow-Credentials": "true",
        "Access-Control-Allow-Headers": (
            "origin, content-type, accept, "
            "authorization, x-xsrf-token, x-request-id"
        ),
    }
    response.headers.extend(headers)


def add_cors_headers(request, response):
    try:
        if request.method != "OPTIONS":
            if request.route and request.route.methods:
                methods = [method for method in request.route.methods]
                _add_cors_headers(response, methods)
    except:
        loguru.logger.warning("Fail add cors headers")



from typing import Any

from sanic.server.websockets.impl import WebsocketImplProtocol

from sanic import Websocket

from web_foundation.workers.io.rt_connection import WriteableObj, RtConnection


class WriteableWs(WriteableObj):
    obj: WebsocketImplProtocol

    def __init__(self, obj: WebsocketImplProtocol):
        super().__init__(obj)

    async def write(self, message: Any) -> None:
        await self.obj.send(message)


class WsRtConnection(RtConnection):
    pass



from web_foundation.app.errors.base_error import BaseAppError


class BaseRouterError(BaseAppError):
    base_ex: Exception

    def __init__(self, base_ex: Exception):
        self.base_ex = base_ex


class YamlRouterLoadFileError(BaseRouterError):
    def __str__(self):
        return f"Can't load file with routes, cause: {self.base_ex.__str__()}"



from typing import Generic, Type

from web_foundation.kernel import IMessage
from web_foundation.workers.background.executor.message import BgTask, TaskIMessage
from web_foundation.workers.worker import GenWorker


class Service(Generic[GenWorker]):
    _worker: GenWorker

    async def run_background(self, task: BgTask, *args, **kwargs):
        await self._worker.channel.produce(
            TaskIMessage(task, args=args, kwargs=kwargs)
        )

    @property
    def worker(self) -> GenWorker:
        return self._worker

    @worker.setter
    def worker(self, worker: GenWorker):
        self._worker = worker

    async def emmit_event(self, event: IMessage):
        await self._worker.channel.produce(event)



import logging
import sys

from loguru import logger
from pydantic import BaseModel


class SanicLogMessage:
    def __init__(self, record):
        self.record = record
        self.msg = " ".join([record.name, record.host, record.request, str(record.status)])


class LoggerSettings(BaseModel):
    log_file: str
    rotation_trigger: str  # "500 MB"  "12:00" "1 week" "10 days"
    compression: str  # "zip"
    retention: str


class LogsHandler(logging.Handler):
    def __init__(self, settings: LoggerSettings | None):
        super().__init__()
        self.settings = settings
        self._set_loggers()

    def _set_loggers(self):
        if self.settings:
            logger.add(self.settings.log_file, rotation=self.settings.rotation_trigger,
                       compression=self.settings.compression, retention=self.settings.retention)

    def emit(self, record):
        # Retrieve context where the logging call occurred, this happens to be in the 6th frame upward
        logger_opt = logger.opt(depth=10, exception=record.exc_info)
        try:
            if record.name == "sanic.access":
                record.msg = SanicLogMessage(record).msg
            if hasattr(record, "status"):
                if record.status < 200:
                    logger_opt.debug(record.msg)
                elif (record.status >= 200) and (record.status < 400):
                    logger_opt.info(record.msg)
                elif (record.status >= 400) and (record.status < 500):
                    logger_opt.warning(record.msg)
                elif record.status >= 500:
                    logger_opt.error(record.msg)
            else:
                logger_opt.log(record.levelname, record.getMessage())
        except Exception as exx:
            logger.warning(f"LOGGER {record.name} ERROR {exx}")
            pass


def setup_loggers(settings: LoggerSettings | None):
    for logger_name, logger_obj in logging.root.manager.loggerDict.items():
        logging.getLogger(logger_name).handlers.clear()
        logging.basicConfig(handlers=[LogsHandler(settings)], level=0, force=True)



from __future__ import annotations

import asyncio
import copy
import os
import socket
from functools import partial
from typing import List, Callable, Any, Type

from dependency_injector import providers
from pydantic import BaseModel as PDModel
from sanic import Sanic
from sanic.server.socket import bind_socket

from web_foundation.app.app import AppContainer
from web_foundation.kernel import GenericIMessage, IMessage
from web_foundation.workers.io.http.ext_error_handler import ExtendedErrorHandler
from web_foundation.workers.io.http.routers.ext_router import ExtRouter
from web_foundation.workers.io.rt_connection import RtConnection, RtEventCallback
from web_foundation.workers.worker import Worker


class ServerConfig(PDModel):
    host: str
    port: int


class HttpServer(Worker):
    config: providers.Configuration
    sock: socket.socket
    sanic_app: Sanic
    router: ExtRouter | None
    rt_connections: List[RtConnection]

    def __init__(self, config: providers.Configuration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.config = config

    def _configure(self, sock: socket.socket = None, ext_router: ExtRouter = None):  # type: ignore
        self.sock = sock if sock else self.create_socket(self.config.get("host"), self.config.get("port"))
        self.router = copy.deepcopy(ext_router)
        self.sanic_app = Sanic(self.name, router=self.router)
        # self.sanic_app.extend(built_in_extensions=False)
        self.sanic_app.error_handler = ExtendedErrorHandler()
        self.sanic_app.after_server_stop(self.close)
        self.sanic_app.ctx.channel = self.channel
        self.rt_connections = []
        self._set_sanic_confs()
        if self.router and issubclass(self.router.__class__, ExtRouter):
            self.router.apply_routes(self.sanic_app)

    def _set_sanic_confs(self):
        self.sanic_app.config.SWAGGER_UI_CONFIGURATION = {
            "docExpansion": 'none'
        }

    async def _run(self):
        # await self._init_resources()
        self.sanic_app.register_listener(self._init_resources, "before_server_start")
        # noinspection PyAsyncCall
        self.sanic_app.add_task(self.channel.listen_consume())

    def _startup(self):
        super(HttpServer, self)._startup()
        try:
            self.sanic_app.run(sock=self.sock)
        except KeyboardInterrupt:
            self._close()

    def _close(self, *args, **kwargs):
        self.sock.close()

    def wire_worker(self, container: AppContainer) -> None:  # type: ignore
        if not self.router:
            raise AttributeError(f"{self.__class__} has no attribute router")
        for chain in self.router.chains:
            for method in chain.methods:
                if method.protector:
                    container.wire(modules=[method.protector.__module__])
                container.wire(modules=[method.handler.__module__])

    @staticmethod
    def create_socket(host: str, port: int) -> socket.socket:
        sock = bind_socket(host, port)
        sock.set_inheritable(True)
        return sock

    async def on_event(self, message: GenericIMessage, resolve_callback: RtEventCallback = None):
        await self.broadcast_rt_events(message, resolve_callback)

    async def broadcast_rt_events(self, event: GenericIMessage,
                                  resolve_callback: RtEventCallback):
        await asyncio.gather(*[conn.send_after_call(event, resolve_callback) for conn in self.rt_connections])

    async def accept_rt_connection(self, conn: RtConnection,
                                   on_disconnect_callback: Callable[[HttpServer, RtConnection], Any] | None = None):
        self.rt_connections.append(conn)

        async def _on_rt_close():
            nonlocal self
            nonlocal conn
            nonlocal on_disconnect_callback
            self.rt_connections.remove(conn)
            if on_disconnect_callback:
                await on_disconnect_callback(self, conn)

        await conn.freeze_request(_on_rt_close)


def create_io_workers(config: providers.Configuration,
                      router: ExtRouter,
                      workers_num: int = 1,
                      fast: bool = False,
                      ) -> List[HttpServer]:
    """
    RESERVE 0 WORKER TO CREATE FRONT SERVICE SERVING
    """
    sock = HttpServer.create_socket(config.get("host"),
                                    config.get("port"))
    workers: List[HttpServer] = []
    w_num = workers_num
    if fast and workers_num != 1:
        raise RuntimeError("You cannot use both fast=True and workers=X")
    if fast:
        try:
            w_num = len(os.sched_getaffinity(0))
        except AttributeError:  # no cov
            w_num = os.cpu_count() or 1
    for i in range(w_num):
        worker = HttpServer(config)
        worker.configure(f"web_worker_{i + 1}", ext_router=router, sock=sock)
        workers.append(worker)
    return workers


def subscribe_worker_rt_callback(*workers: HttpServer,
                                 event_type: List[Type[IMessage] | str] | Type[IMessage] | str,
                                 callback: RtEventCallback,
                                 use_nested_classes: bool = False):
    for worker in workers:
        worker.channel.add_event_listener(event_type, partial(worker.on_event, resolve_callback=callback),
                                          use_nested_classes=use_nested_classes)



from distutils.core import setup
import os

# def package_files(directory):
#     paths = []
#     for (path, directories, filenames) in os.walk(directory):
#         for filename in filenames:
#             paths.append(os.path.join('..', path, filename))
#     return paths
#
#
# extra_files = package_files('./web_foundation')
# print(extra_files)

DEPTH = 20

setup(
    name='web-foundation',  # How you named your package folder (MyLib)
    packages=['web_foundation'],  # Chose the same as "name"
    package_data={
        'web_foundation': [('*/' * i).rstrip("/") for i in range(1, DEPTH)],
    },
    version='0.0.6.23',  # Start with a small number and increase it with every change you make
    license='MIT',  # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    description='Python microprocessors web-foundation',  # Give a short description about your library
    author='Yaroha',  # Type in your name
    author_email='yaroher2442@gmail.com',  # Type in your E-Mail
    keywords=['SOME', 'MEANINGFULL', 'KEYWORDS'],  # Keywords that define your package best
    install_requires=[  # I get to this in a second
        "sanic==22.6.0",
        "dependency-injector>=4.0,<5.0",
        "sanic-ext==22.6.1",
        "tortoise-orm==0.19.1",
        "asyncpg==0.26.0",
        "pydantic==1.9.1",
        "aioprocessing==2.0.0",
        "pydantic==1.9.1",
        "dill==0.3.5.1",
        "pyee==9.0.4",
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
    ],
)



import re
from datetime import datetime
from typing import Union, List
from datetime import time as DTime

# ------------------------------------------------STR_TO----------------------------------------------------------------
from web_foundation.app.errors.application.application import InconsistencyError


def str_to_list(string: str) -> Union[List[str], None]:
    if not ",":
        return None
    return string.split(",")


def str_to_bool(string: str) -> Union[bool, None]:
    try:
        if string.lower() == "false":
            return False
        elif string.lower() == "true":
            return True
        else:
            return None
    except Exception as e:
        return None


# ------------------------------------------------VALIDATE--------------------------------------------------------------


def validate_date(date: str, check_gt_now=True) -> str:
    try:
        if "." in date:
            _date = datetime.strptime(date, "%d.%m.%Y")
            date = _date.isoformat()
        else:
            _date = datetime.fromisoformat(date)
    except ValueError as exception:
        raise InconsistencyError(exception)
    if check_gt_now and datetime.now() > _date:
        raise InconsistencyError(message="Incorrect date")
    return date


def validate_time(str_time: str) -> str:
    try:
        t = DTime.fromisoformat(str_time)
    except ValueError as exception:
        raise InconsistencyError(exception)
    return str_time


def validate_datetime(date_time: str, check_gt_now=False) -> str:  # TODO add timezones
    format_types = ["%d", "%m", "%Y", "%H", "%M", "%S"]
    try:
        if "." in date_time and date_time.count(".") > 1:
            date_nums = [ddd for ddd in re.findall(r"(\d*)", date_time) if ddd]
            _date = datetime.strptime(' '.join(date_nums), ' '.join(format_types[:len(date_nums)]))
        else:
            _date = datetime.fromisoformat(date_time)
        date = _date.astimezone().isoformat()
    except ValueError as exception:
        raise InconsistencyError(exception)
    if check_gt_now and datetime.now() > _date:
        raise InconsistencyError(message="Incorrect datetime")
    return date



import asyncio
from dataclasses import dataclass
from typing import Coroutine, Callable, Dict, Type, List
from typing import TypeVar

import loguru
from aioprocessing import AioEvent, AioLock, AioQueue  # type: ignore
from loguru import logger


class IMessage:
    message_type: str = "__all__ "
    index: int
    sender: str

    def __init__(self):
        self.index = 0
        self.sender = "None"

    def __str__(self):
        return f"IMessage(id={self.index}, sender={self.sender})"


GenericIMessage = TypeVar("GenericIMessage", bound=IMessage, contravariant=True)


@dataclass
class IsolatePipes:
    queue: AioQueue
    event: AioEvent
    lock: AioLock

    async def write(self, msg: GenericIMessage):
        await self.queue.coro_put(msg)

    async def read(self) -> GenericIMessage:
        return await self.queue.coro_get()

    def empty(self):
        return self.queue.empty()


EventListener = Callable[[GenericIMessage], Coroutine]


class IChannel:
    read_timeout = 0.01
    pipes: IsolatePipes
    worker_name: str
    debug: bool
    consume_pipe: IsolatePipes
    produce_pipe: IsolatePipes
    _listeners: Dict[str, List[EventListener]]

    def __init__(self, isolate_name: str, debug: bool = False):
        self.worker_name = isolate_name
        self.debug = debug
        self._listeners = {}
        self.produce_pipe = IsolatePipes(AioQueue(), AioEvent(), AioLock())
        self.consume_pipe = IsolatePipes(AioQueue(), AioEvent(), AioLock())

    def __str__(self):
        return f"IChannel(name={self.worker_name})"

    async def produce(self, msg: IMessage):
        msg.sender = self.worker_name
        await self.produce_pipe.write(msg)

    async def listen_produce(self, callback: Callable[[GenericIMessage], Coroutine]):
        while True:
            if self.produce_pipe.empty():
                await asyncio.sleep(self.read_timeout)
            r: GenericIMessage = await self.produce_pipe.read()
            if self.debug:
                logger.info(
                    f"Channel {self.worker_name} send message: {r.message_type}")
            await callback(r)
            await asyncio.sleep(0.01)

    async def listen_consume(self):
        while True:
            if not self.consume_pipe.empty():
                r: IMessage = await self.consume_pipe.read()
                if self.debug:
                    logger.info(
                        f"Channel {self.worker_name} receive message: {r.message_type}")
                callbacks = self._listeners.get(r.message_type)
                if not callbacks:
                    continue
                for callback in callbacks:
                    asyncio.create_task(callback(r))
            await asyncio.sleep(0.01)

    def _add_event_listener(self, event_type: Type[IMessage] | str, callback: EventListener,
                            use_nested_classes: bool = False):
        if use_nested_classes and isinstance(event_type, str):
            raise AttributeError("Can't add_event_listener with use_nested_classes")

        def _add(event_name: str):
            nonlocal self
            nonlocal callback
            if event_name not in self._listeners:
                self._listeners[event_name] = []
            self._listeners[event_name].append(callback)

        if isinstance(event_type, str):
            _add(event_type)
        else:
            def _raise(cls_type):
                if not hasattr(cls_type, "message_type"):
                    raise AttributeError("Can't register listener cause message_type not found")

            if use_nested_classes:
                for cls in event_type.__subclasses__():
                    _raise(cls)
                    _add(cls.message_type)
            else:
                _raise(event_type)
                _add(event_type.message_type)

    def add_event_listener(self, event_type: List[Type[IMessage] | str] | Type[IMessage] | str,
                           callback: EventListener, use_nested_classes: bool = False):
        if isinstance(event_type, list):
            for i in event_type:
                self._add_event_listener(i, callback, use_nested_classes)
        else:
            self._add_event_listener(event_type, callback, use_nested_classes)

    def remove_event_listeners(self, event_type: Type[IMessage] | str):
        if isinstance(event_type, str):
            self._listeners.pop(event_type)
        else:
            self._listeners.pop(event_type.message_type)



from datetime import datetime
from datetime import date as Ddate
from datetime import time as DTime

from web_foundation.app.errors.application.application import InconsistencyError


def string_from_db_date(db_date: datetime | Ddate | DTime | str, iso_date_format=False) -> str:
    v = db_date
    if isinstance(db_date, datetime):
        if iso_date_format:
            v = db_date.replace(microsecond=0).astimezone().isoformat()
        else:
            v = db_date.astimezone().strftime("%d.%m.%Y, %H:%M:%S %z")
    elif isinstance(db_date, Ddate):
        if iso_date_format:
            v = db_date.isoformat()
        else:
            v = db_date.strftime("%d.%m.%Y")
    elif isinstance(db_date, DTime):
        v = db_date.replace(microsecond=0).isoformat()
    return str(v)


def integrity_error_format(exception) -> InconsistencyError:
    str_exception = str(exception)

    if "table" in str_exception:  # tortoise.exceptions.IntegrityError: insert or update on table "issue" violates foreign key constraint "issue_service_id_fkey"
        start_tablename = str_exception.rfind("table") + 7
        end_tablename = str_exception.rfind(".", start_tablename) - 1
        model_name = str_exception[start_tablename:end_tablename].capitalize()
        return InconsistencyError(
            message=f"Related model {model_name} not exist")
    elif str_exception.startswith("null"):
        start_field_name = str_exception.find('"') + 1
        end_field_name = str_exception.rfind('"')
        return InconsistencyError(message=f"{str_exception[start_field_name:end_field_name]} can't be null")
    elif "already exist" in str_exception:
        return InconsistencyError(message="Already exist")
    else:
        return InconsistencyError(exception)



import hashlib
import logging
import random
import string
import sys
from base64 import b64encode, b64decode
from typing import Iterable
from uuid import uuid4


# from Crypto.Cipher import AES
# from Crypto.Random import get_random_bytes


class BaseCrypto:
    class DataStruct:
        pass

    def __init__(self):
        pass

    @staticmethod
    def encrypt_password(passw: str, salt_len: int = 10) -> Iterable[str]:
        salt = ''.join(random.choices(string.ascii_uppercase + string.digits, k=salt_len))
        crypted_pass = hashlib.sha256((passw + salt).encode(sys.getdefaultencoding())).hexdigest()
        return crypted_pass, salt

    @staticmethod
    def verify_password(raw_pass: str, db_pass: str, salt: str):
        crypted_pass = hashlib.sha256((raw_pass + salt).encode(sys.getdefaultencoding())).hexdigest()
        return db_pass == crypted_pass

    def encrypt(self, plain_text: str) -> DataStruct:
        raise NotImplementedError()

    def decrypt(self, aes_struct: DataStruct) -> str:
        raise NotImplementedError()


class AESCrypto(BaseCrypto):
    class DataStruct:
        __slots__ = ["cipher_text", "salt", "nonce", "tag"]

        def __init__(self, cipher_text, salt, nonce, tag):
            self.cipher_text = cipher_text
            self.salt = salt
            self.nonce = nonce
            self.tag = tag

        def decode(self):
            for var, val in self.__dict__().items():
                setattr(self, var, b64decode(val))

        def encode(self):
            encoding = sys.getdefaultencoding()
            for var, val in self.__dict__().items():
                setattr(self, var, b64encode(val).decode(encoding))

        def __dict__(self):
            return {"cipher_text": self.cipher_text,
                    "salt": self.salt,
                    "nonce": self.nonce,
                    "tag": self.tag}

        def __str__(self):
            return "\n".join([f"{k} : {v}" for k, v in self.__dict__().items()])

    _data_password: str
    _n_factor: int
    _block_size: int
    _threads_num: int
    _key_len: int

    def __init__(self, secret: str):
        super(AESCrypto, self).__init__()
        self._data_password = secret
        self._n_factor = 2 ** 14
        self._block_size = 8
        self._threads_num = 1
        self._key_len = 32

    @property
    def key_len(self):
        return self._key_len



import logging

from loguru import logger


class SanicLogMessage:
    def __init__(self, record):
        self.record = record
        self.msg = " ".join([record.worker_name, record.host, record.request, str(record.status)])


class LogsHandlerExt(logging.Handler):
    def emit(self, record):
        # Retrieve context where the logging call occurred, this happens to be in the 6th frame upward
        logger_opt = logger.opt(depth=10, exception=record.exc_info)
        try:
            if record.name == "sanic.access":
                record.msg = SanicLogMessage(record).msg
            if hasattr(record, "status"):
                if record.status < 200:
                    logger_opt.debug(record.msg)
                elif (record.status >= 200) and (record.status < 400):
                    logger_opt.info(record.msg)
                elif (record.status >= 400) and (record.status < 500):
                    logger_opt.warning(record.msg)
                elif record.status >= 500:
                    logger_opt.error(record.msg)
            else:
                logger_opt.log(record.levelname, record.getMessage())
        except Exception as exx:
            logger.warning(f"LOGGER {record.name} ERROR {exx}")
            pass

    @staticmethod
    def set_logger():
        for logger_name, logger_obj in logging.root.manager.loggerDict.items():
            logging.getLogger(logger_name).handlers.clear()
            logging.basicConfig(handlers=[LogsHandlerExt()], level=0)
